package parser;

import java.util.ArrayList;
import java.util.List;

import liner.LinerUtil;
import liner.ScriptForm;
import util.Util;
import domain.Action;
import domain.BlankLine;
import domain.NewScene;
import domain.PreScript;
import domain.ScriptObject;
import domain.Dialogue;
import domain.Scene;
import domain.Script;
import domain.ScriptObjectType;
import domain.ShotChange;
import domain.Transition;

/**
 * The parser
 * 
 * @author David Smith
 *
 */
public class Parser {
	
	/**
	 * converts a script as an array of text into a script object
	 * @param text the array of text
	 * @param scriptName the name of the script
	 * @return the script object
	 */
	public static Script textToScript(ArrayList<String> text, String scriptName) {
		Script script = new Script(scriptName, text.size());
		final int LAST_LINE = text.size()-1;
		int i = 0;
		Scene curScene = null;
		ShotChange shot = null;
		NewScene newScene = null;
		Integer sceneNum = 0;
		BlankLine blank = new BlankLine();
		script.addSceneObject(blank, null);
		PreScript pre = new PreScript();
		script.addSceneObject(pre, null);
		
		while (i <= LAST_LINE) {
			if (Util.isScene(text.get(i))) { 
				sceneNum++;
				curScene = Util.parseSceneString(text.get(i));
				curScene.setId(sceneNum);
				script.getScenes().add(curScene);

				newScene = new NewScene();
				newScene.setScriptLine(i);
				newScene.setScene(sceneNum);

				curScene.addSceneObject(script.addSceneObject(newScene, sceneNum));
				script.addLine(i, newScene.getIndex());
				i++;
			}
			else if (curScene==null) {
				if (i==LAST_LINE) {
					throw new RuntimeException("No scenes found in script. Looking for INT/EXT markers.");
				}
				else {
					script.addLine(i, pre.getIndex());
					i++;
				}
			}
			else if (Util.isCharacter(text, i, LAST_LINE))
			{
				Dialogue dia = new Dialogue();
				dia.setScriptLine(i);
				dia.setScene(sceneNum);
				
				String name = text.get(i).trim();
				String view = null;
				if (name.contains("(")) {
					String[] nameSplit = name.split("\\(");
					name = nameSplit[0].trim();
					view = nameSplit[1].replace(")", "");
					dia.addView(view);
				}
				ScriptObject c = script.addObject(name, sceneNum);
				c.setType(ScriptObjectType.Character);
				curScene.addScriptObject(c.getIndex());
				dia.setCharacter(c.getIndex());
				
				curScene.addSceneObject(script.addSceneObject(dia, sceneNum));
				script.addLine(i, dia.getIndex());
				i++;
				if (text.get(i).contains("(")) {
					StringBuilder view2 = new StringBuilder();
					view2.append(text.get(i).replace("(", "").replace(")", "").trim());
					while (!text.get(i).contains(")"))  {
						script.addLine(i, dia.getIndex());
						i++;
						view2.append(" ");
						view2.append(text.get(i).replace(")","").trim());
					}
					dia.addView(view2.toString());
					script.addLine(i, dia.getIndex());
					i++;
				}
				StringBuilder content = new StringBuilder();
				boolean nextBlank = false;
				while (!nextBlank) {
					if (text.get(i).contains("(more)")) {
						script.addLine(i, dia.getIndex());
						i++;
					}
					content.append(text.get(i).trim());
					script.addLine(i, dia.getIndex());
					i++;
					nextBlank = Util.isBlank(text.get(i)); 
					if (!nextBlank)
						content.append(" ");
				};
				dia.setContent(content.toString());
			}
			else if (i>0 && i<LAST_LINE && 
					Util.isTransition(text.get(i), text.get(i-1), text.get(i+1)))
			{
				Transition tran = new Transition();
				tran.setScriptLine(i);
				tran.setContent(text.get(i).trim());
				tran.setScene(sceneNum);
				curScene.addSceneObject(script.addSceneObject(tran, sceneNum));
				script.addLine(i, tran.getIndex());
				i++;
			}
			else if (Util.isBlank(text.get(i))){
				script.addLine(i, blank.getIndex());
				i++;
			}
			else {
				Action action = new Action();
				action.setScriptLine(i);
				action.setScene(sceneNum);
				curScene.addSceneObject(script.addSceneObject(action, sceneNum));
				StringBuilder content = new StringBuilder();
				boolean done = false;
				while (!done) {
					content.append(text.get(i).trim());
					
					script.addLine(i, action.getIndex());
					i++;
					done = (i >= LAST_LINE) || Util.isBlank(text.get(i)) || Util.isScene(text.get(i));
					if (!done)
						content.append(" ");
				};
				
				//find important words and add to script as objects
				ArrayList<String> importantWords = Util.getUpperWords(content.toString());
				for (String word: importantWords) {
					ScriptObject newObject = script.addObject(word, sceneNum);
					curScene.addScriptObject(newObject.getIndex());
					action.addObject(newObject.getIndex());
				}
				
				action.setContent(content.toString());
			}
		}
		return script;
	}
	
	/**
	 * Checks if the script object is a camera object
	 * @param obj the object being checked
	 * @return true if the object is a camera object
	 */
	public static boolean isCamera(ScriptObject obj) {
		if (obj.getName().indexOf("POV") != -1
				|| obj.getName().indexOf("PANS") != -1
				|| obj.getName().indexOf("INTERCUT") != -1)
			return true;
		else
			return false;
	}

	/**
	 * Checks if the object is a group object
	 * @param group the object being checked
	 * @param objs the list of all objects
	 * @return true if the object is a script object
	 */
	public static boolean isGroup(ScriptObject group, List<ScriptObject> objs) {
		int count = 0;
		for (ScriptObject obj: objs) {
			if (group.getName().contains(obj.getName().trim()) 
					&& obj.getType().equals(ScriptObjectType.Character))
				count++;
		}
		return count > 1;
	}

}
