package util;
import java.util.ArrayList;
import java.util.Random;

import domain.ColorValue;
import domain.IntExt;
import domain.Scene;

/**
 * 
 * A suite of utilities for reading and processing script strings
 * 
 * @author David Smith
 *
 */
public class Util {
	
	public static Random random = new Random();
	
	/**
	 * Trims the lead spaces off string
	 * @param s the string to trim
	 * @return the trimmed string
	 */
	public static String trimLead(String s) {
	    return s.replaceAll("^\\s+", "");
	}
	
	/**
	 * Checks if the string starts and ends with the same number
	 * @param line the string tested
	 * @return true if starts with first and last numbers
	 */
	public static boolean isFirstAndLastNum(String line) {
		String[] words = line.replaceAll("^\\s+", "").trim().split(" ");
		return (words[0].equals(words[words.length-1])) && isNumeric(words[0]);
	}
	
	/**
	 * removes the first and last number from the string by replacing with blank spaces
	 * @param line the input string
	 * @return the string without numbers
	 */
	public static String removeFirstAndLastNum(String line) {
		String[] words = line.replaceAll("^\\s+", "").trim().split(" ");
		if (!words[0].startsWith("INT") 
				&& !words[0].startsWith("EXT")
				&& isNumeric(words[0].substring(0,1))) {
			line = line.replaceAll(words[0], " ");
		}
		return line;
	}
	
	/**
	 * Tests if line is a scene header
	 * @param inLine the string being tested
	 * @return true if the line is a scene header
	 */
	public static boolean isScene(String inLine) {
		String line = inLine.replaceAll("^\\s+", "");
		boolean b = (line.startsWith("INT ") || 
				line.startsWith("INT. ") ||
				line.startsWith("INTERIOR") ||
				line.startsWith("EXT ") ||
				line.startsWith("EXT.") ||
				line.startsWith("EXTERIOR") ||
				line.indexOf("INT.") != -1 ||
				line.indexOf("EXT.") != -1) && isMostlyUpper(line, .9);
		return b || isFirstAndLastNum(line);
	}
	
	/**
	 * tests if the line blank
	 * @param line the line being tested
	 * @return true if the line is blank
	 */
	public static boolean isBlank(String line) {
		return line == null || line.trim().isEmpty();
	}
	
	/**
	 * tests if a line is all upper case letters
	 * @param line the line being tested
	 * @return true if the line all upper case
	 */
	public static boolean isAllUpper(String line) {
		boolean hasLower = false;
		boolean hasUpper = false;
		for (char c: line.toCharArray()) {
			if (Character.isLowerCase(c))
				hasLower=true;
			if (Character.isUpperCase(c))
				hasUpper=true;
		}
		return hasUpper && !hasLower;
	}
	
	/**
	 * Tests if a line is mostly uppercase letters based on a threshold percentage
	 * @param line the line being tested
	 * @param threshold the threshold for "mostly", between 0 and 1
	 * @return true if the line is mostly upper
	 */
	public static boolean isMostlyUpper(String line, double threshold) {
		double uppers = 0;
		double lowers = 0;
		for (char c: line.toCharArray()) {
			if (Character.isLowerCase(c))
				lowers++;
			if (Character.isUpperCase(c))
				uppers++;
		}
		if (uppers+lowers == 0)
			return false;
		double percentage = uppers/(uppers+lowers);
		return (percentage >= threshold);
	}
	
	/**
	 * Gets all the upper words in a line
	 * @param line the line to search the words for
	 * @return an array of the upper words
	 */
	public static ArrayList<String> getUpperWords(String line) {
		ArrayList<String> upperWords = new ArrayList<String>();
		String[] allWords = line.split(" ");
		boolean lastUpper = false;
		for (int i=0;i<allWords.length;i++) {
			if(isAllUpper(allWords[i])
					&& !allWords[i].equals("I")
					&& !allWords[i].equals("A")) {
				if (lastUpper) 
					upperWords.set(upperWords.size()-1, upperWords.get(upperWords.size()-1) + " " + allWords[i].trim());
				else
					upperWords.add(allWords[i].trim());
				lastUpper = true;
			}
			else
				lastUpper = false;
		}
		return upperWords;
	}
	
	/**
	 * Checks if the current line is a character, the first part of a dialogue
	 * @param text all the lines of text
	 * @param index the current index of the text being checked
	 * @param lastLine the index for the last line of the text
	 * @return true if the line represents a character name
	 */
	public static boolean isCharacter(ArrayList<String> text, int index, int lastLine) {

		return index>0 && index<lastLine && isAllUpper(text.get(index)) &&  isBlank(text.get(index-1)) 
				&& !isBlank(text.get(index+1)) && !isScene(text.get(index+1));
	}
	
	/**
	 * Checks if the line is a transition
	 * @param line the line being checked
	 * @param prev the previous line
	 * @param next the next line
	 * @return true if the line is a transition
	 */
	public static boolean isTransition(String line, String prev, String next) {
		return isAllUpper(line) && isBlank(prev) && isBlank(next) && line.trim().endsWith(":");
	}
	
	/**
	 * Removes any strange characters from the end of a string, including periods and commas
	 * Removes all paranthesis
	 * @param in The line being cleaned
	 * @return the cleaned line
	 */
	public static String clean(String in) {
		String out = in;
		if (in.endsWith(".") 
				|| in.endsWith(",")
				|| in.endsWith("!")
				|| in.endsWith("?"))
			out = in.substring(0, in.length()-1);
		out = out.replaceAll("[()]", "");
		return out;
	}
	
	/**
	 * Checks to see if the string is the when part of a scene header
	 * @param line the line being checked
	 * @return true if the line is part of a scene header
	 */
	public static boolean isWhen(String line) {
		line = line.trim();
		return (line.equals("DAY") ||
				line.equals("NIGHT") ||
				line.equals("DAWN") ||
				line.equals("MORNING") ||
				line.equals("EVENING") ||
				line.equals("DUSK") || 
				line.equals("LATER"));
	}
	
	/**
	 * Parses the line into a scene object
	 * @precondition the line must be a scene header
	 * @param line the line being parsed
	 * @return a new scene
	 */
	public static Scene parseSceneString(String line) {
		Scene scene = new Scene();
		line = removeFirstAndLastNum(line);
		String[] whereSplit = line.trim().split(" ", 2);
		while (!whereSplit[0].startsWith("INT") 
				&& !whereSplit[0].startsWith("EXT") 
				&& whereSplit.length == 2) {
			whereSplit = whereSplit[1].split(" ", 2);
		}
		if (whereSplit.length < 2) {
			scene.setWhere("");
			scene.setLocation(line);
			scene.setWhen("");
		}
		else {
			scene.setWhere(whereSplit[0]);
			String next = whereSplit[1];
			next = next.replaceAll("--", "-");
			if (next.contains("-")) {
				String[] whenSplit = next.split("\\-");
				String when = whenSplit[whenSplit.length-1].trim();
				if (isWhen(when))
					scene.setWhen(when);
				else
					scene.setWhen("");
				String location = whenSplit[0].trim();
				for (int i=1; i<whenSplit.length; i++)
					if(!isWhen(whenSplit[i]))
						location = location + " - " + whenSplit[i].trim();
				scene.setLocation(location);
			}
			else {
				scene.setLocation(next.trim());
				scene.setWhen("");
			}
		}
		return scene;
	}
	
	/**
	 * Increases the scene number
	 * @param ctr the input counter
	 * @return the new scene
	 */
	public static String increaseShot(String ctr) {
		if (ctr == null)
			return "A";
		int len = ctr.length();
		char[] num = new char[len];
		for (int i=0; i< len; i++)
			num[i] = ctr.charAt(i);
		StringBuilder newShot = new StringBuilder();
		if (num[len-1] < 'Z') {
			num[len-1]++;
		}
		else {
			int pos = len -1;
			while (num[pos] == 'Z') {
				num[pos] = 'A';
				pos--;
				if (pos < 0) {
					newShot.append("A");
					break;
				}
			}
			if (pos>=0 && num[pos] < 'Z')
				num[pos]++;
		}
		for (int i=0; i<len; i++)
			newShot.append(num[i]);
		return newShot.toString();
	}

	/**
	 * Checks if a string is all numbers
	 * @param str the string being checked
	 * @return true if the string is a number
	 */
	public static boolean isNumeric(String str)
	{
		return str.matches("-?\\d+(\\.\\d+)?");
	}
	
	/**
	 * Generates a random color
	 * @return a random color
	 */
	public static ColorValue randomColor() {
		ColorValue color = new ColorValue(random.nextFloat(), random.nextFloat(), random.nextFloat());
		return color;
	}
	
	/**
	 * Strips all extra characters and symbols out
	 * @param s the string to strip down
	 * @return the stripped down string
	 */
	public static String strippedDown(String s) {
		return s.toLowerCase()
				.replace(",", "")
				.replace(".", "")
				.replace("!", "")
				.replaceAll("-", "")
				.replace("?", "")
				.replace("the ", "")
				.replace("(v.o.)", "")
				.replace("\\(", "")
				.replace("\\)", "")
				.trim();
				
	}

	/**
	 * Returns if a string is interior or exterior
	 * @param where the where string
	 * @return the interior exterior enum
	 */
	public static IntExt getIntExt(String where) {
		if (where.toUpperCase().contains("INT"))
			return IntExt.Int;
		else if (where.toUpperCase().contains("EXT"))
			return IntExt.Ext;
		else
			return IntExt.none;
	}
}
